//
//  n80em.m
//
//  Copyright (C) 1993-2019 Hiroshi Kanemoto.
//    Released under the MIT license
//    http://opensource.org/licenses/mit-license.php
//
//     Created:    2012/08/22    H.Kanemoto
//    Modified:    2016/04/05    H.Kanemoto    ported from MaxOS X
//    Modified:    2019/05/12    H.Kanemoto    support METAL
//
#import "n80em.h"
#import "n80sound.h"
#import "n80graph.h"
#import "n80keyboard.h"
#import <stdatomic.h>
#include <mach/mach_time.h>
#include <AudioUnit/AudioUnit.h>
//-----------------------------------------------------------------------------
//    Emulator C Functions
//-----------------------------------------------------------------------------
#define NOTDONE    0
#define DONE    1
#define STDSIZE    (32576)
#define MAXSIZE    (32768)
#define SP_VAL    (0x7f3e)
#define PC_VAL    (0xff3d)

export N80LX        n80lx;
static n80em        *emu;
//-----------------------------------------------------------------------------
//    export "C" functions of bridge Objective-C for n80lx emulator module
//-----------------------------------------------------------------------------
static void _n80_beep_on(void) {
    [emu.n80snd beep_on];
}
static void _n80_beep_off(void) {
    [emu.n80snd beep_off];
}
static void _n80_sound_on(USHORT freq, int ch) {
    [emu.n80snd sound_on:freq ch:(int)ch];
}
static void _n80_sound_off(void) {
    [emu.n80snd sound_off];
}
static void _n80_chg_bltfunc(int mode) {
    // mode: 0 = 40x20, 1 = 40x25, 2 = 80x20, 3 = 80x25
    [emu.n80gr chg_bltfunc:mode];
}
static void _n80_chg_palette(bool interrace) {
    [emu.n80gr chg_palette:interrace];
}
static void _n80_vram_enter(void) {
    [emu.n80gr lock];
}
static void _n80_vram_leave(void) {
    n80lx.vram_sync = 0;
    [emu.n80gr unlock];
}
export void _n80_init_snd(void) {
    n80device.snd.beep_on      = _n80_beep_on;
    n80device.snd.beep_off     = _n80_beep_off;
    n80device.snd.sound_on     = _n80_sound_on;
    n80device.snd.sound_off    = _n80_sound_off;
}
export void _n80_init_blt(void) {
    n80device.blt.chg_crt_mode = _n80_chg_bltfunc;
    n80device.blt.chg_palette  = _n80_chg_palette;
}
//-----------------------------------------------------------------------------
//    supplement function
//-----------------------------------------------------------------------------
static double system_tick_time(void) {
    struct timeval time;
    gettimeofday(&time, NULL);
    return time.tv_sec + (double) time.tv_usec*1e-6;
}
//-----------------------------------------------------------------------------
//    Emulator Objective-C Class
//-----------------------------------------------------------------------------
@implementation n80em

-(id)initWithDma:(DblBuffBmpImage*)bmp BootROM:(BYTE)mode Penalty:(int)penalty Interrace:(bool)interrace {
    if (self = [super init]) {
        emu = self;
        // init member variables
        self->dblbmp = bmp;
        self.running  = false;
        self.terminating = false;    // leave pause before terminating
        self.haltcallback = nil;
        
        // N80 structure clean
        memset(&n80lx, 0, sizeof(n80lx));
        
        // Z80 Initialize
        __z80_set_external_method(&n80lx.cpu, _n80_mbread, _n80_mbwrite, _n80_ioread, _n80_iowrite);
        
        // PC-8001 Device Initialize
        _n80_device_initialize(&n80device, _n80_vram_enter, _n80_vram_leave);
        
        // load font file
        if (![self loadFontFile])
            return nil;
        
        // load rom file
        if (![self loadROMFile])
            return nil;
        
        pauseZ80sync = [[NSConditionLock alloc] initWithCondition:NOTDONE];

        // create/initialize sound module
        self->sound = [[n80sound alloc] initWithN80:self];
        // create/initialize graph module
        self->graph = [[n80graph alloc] initWithN80:self];
        // create/initialize keyboard module
        self->keyboard = [[n80keyboard alloc] initWithN80:self];
        
        _n80_chg_palette(interrace);
    }
    return self;
}

-(void)dealloc {
    NSLog(@"n80em dealloc");
    [self reqDispose];
}

-(N80LX const *)getN80 {
    return &n80lx;
}

-(n80sound const *)getN80Sound {
    return self->sound;
}

-(n80graph const *)getN80Graph {
    return self->graph;
}

-(n80keyboard const *)getN80Keyboard {
    return self->keyboard;
}

-(DblBuffBmpImage const *)getBmpimg {
    return self->dblbmp;
}

-(int)getPenalty {
    return n80device.dmac.dma_penalty_base;
}

-(void)setPenalty:(int)DMAPenalty {
    n80device.dmac.dma_penalty_base = DMAPenalty;
}

-(void)suspend {
    [self.n80snd suspend];
    [self.n80gr suspend];
    self.running  = false;
    self->state = 0;
}

-(void)resume {
    if ([self.n80snd resume]) {
        [self start];
    }
}

- (void)start {
    self.running  = true;
    self.pause = false;
    self->state = 0;
    [self.n80gr start];
}
- (void)reqDispose {
    atomic_store(&self->request_cpu_signal, 1);
    if (self.running) {
        self.running = false;
    }
    // stop sound unit
    [self.n80snd reqDispose];
    if (n80device.cmt._buff != NULL) {
        free(n80device.cmt._buff);
        n80device.cmt._buff = NULL;
    }
    // release graph unit
    [self.n80gr reqDispose];
    self->dblbmp = nil;
}
- (void)reqReset {
    int expected = 0;
    while (!atomic_compare_exchange_strong(&self->request_cpu_signal, &expected, 2)) {
        expected = 0;
    }
}
- (void)reqPCGToggle {
    n80device.pcg_mode ^= 0x01;
    n80device.attr.invalid();
}
#if TARGET_OS_OSX
- (void)reqOpenFile {
    [pauseZ80sync lock];

    FILE *fp = NULL;
    BOOL isN80File = false;
    NSOpenPanel *openPanel = [NSOpenPanel openPanel];
    NSArray *allowedFileTypes = [NSArray arrayWithObjects:@"N80",@"'N80 Memory Image File'",@"CMT",@"'Casset Image File'",nil];
    [openPanel setAllowedFileTypes:allowedFileTypes];
    NSInteger pressedButton = [openPanel runModal];
    if (pressedButton == NSModalResponseOK) {
        NSString *path = [[openPanel URL] path];
        isN80File = ([path hasSuffix:@"N80"] || [path hasSuffix:@"n80"]);
        fp = fopen([path fileSystemRepresentation],"r");
    }
    if (fp != NULL) {
        fseek(fp, 0, SEEK_END);
        long fsiz = ftell(fp);
        rewind(fp);

        if (isN80File) {    // N80 File
            USHORT spval;
            
            if (fsiz >= STDSIZE) {
                fsiz = (fsiz > MAXSIZE) ? MAXSIZE : fsiz;
                fread(n80device.mem.map._ram, 1, fsiz, fp);
                fseek(fp, SP_VAL, SEEK_SET);
                fread(&spval, 2, 1, fp);
                n80lx.cpu.r.x.PC = PC_VAL;
                n80lx.cpu.r.x.SP = spval;
            }
        } else {            // CMT File
            if (n80device.cmt._buff != NULL) {
                free(n80device.cmt._buff);
                n80device.cmt._buff = NULL;
            }
            if (fsiz > 0) {
                n80device.cmt._buff = malloc(fsiz);
                size_t len = fread(n80device.cmt._buff, 1, fsiz, fp);
                n80device.cmt._size  = (int) len;
                n80device.cmt._index = 0;
            }
        }
        fclose(fp);
    }
    [pauseZ80sync unlock];
}
#else
- (void)reqOpenFile:(bytecoder*)data isN80:(BOOL)isN80File {
	[pauseZ80sync lock];
	int expected = 0;
	while (!atomic_compare_exchange_strong(&self->request_cpu_signal, &expected, 3)) {
		expected = 0;
	}
	//OSAtomicCompareAndSwap32(0, 3, &n80lx.request_disp_signal);
	
	if (isN80File) {	// N80 File
		USHORT spval;
		
		int fsiz = data.Length;
		if (fsiz >= STDSIZE) {
			fsiz = (fsiz > MAXSIZE) ? MAXSIZE : fsiz;
			[data moveptrto:0];
			[data pull:&n80device.mem.map._ram[0] Length:fsiz];
			[data moveptrto:SP_VAL];
			[data pull_le_short:&spval];
			n80lx.cpu.r.x.PC = PC_VAL;
			n80lx.cpu.r.x.SP = spval;
		}
	} else {			// CMT File
		if (n80device.cmt._buff != NULL) {
			free(n80device.cmt._buff);
			n80device.cmt._buff = NULL;
		}
		n80device.cmt._buff = malloc(data.Length);
		[data moveptrto:0];
		[data pull:n80device.cmt._buff Length:data.Length];
		n80device.cmt._size  = (int) data.Length;
		n80device.cmt._index = 0;
	}
	[pauseZ80sync unlock];
}
#endif
//-----------------------------------------------------------------------------
//    Initialize Emulator
//-----------------------------------------------------------------------------
-(BOOL)loadFontFile {
    if ([self loadResFile:@"PC-8001.FON" Size:1024*4 Buffer:n80device.font]) {
        // copy font rom data to PCG ram
        for (int i = 0; i < 2048; ++i)
            n80device.pcg[i] = n80device.font[i];
        return true;
    }
    return false;
}

-(BOOL)loadROMFile {
    return [self loadResFile:@"PC-8001.ROM" Size:1024*24 Buffer:&n80device.mem.map._rom[0]];
}

-(BOOL)loadResFile:(NSString*)filename Size:(int)size Buffer:(LPBYTE)buffer {
    // find rom file from App local path
    NSString *defPath = [[NSBundle mainBundle] pathForResource:filename ofType:nil];
    FILE *fp = NULL;
    fp = fopen([defPath fileSystemRepresentation], "r");
    if (fp == NULL)
        return false;
    
    // read data from resource file
    size_t len = fread(buffer, 1, size, fp);
    fclose(fp);
    if (len != size)
        return false;
    return true;
}
#if TARGET_OS_IPHONE
+(bytecoder*)loadFile:(NSString*)fullpath {
	FILE *fp = NULL;
	fp = fopen([fullpath fileSystemRepresentation], "r");
	if (fp == NULL)
		return nil;
	fpos_t fsize = 0;
	fseek(fp, 0, SEEK_END);
	fgetpos(fp, &fsize);
	fseek(fp, 0, SEEK_SET);
	UInt8 buff[fsize];
	size_t len = fread(buff, 1, (size_t)fsize, fp);
	fclose(fp);
	
	bytecoder *data = [[bytecoder alloc] init];
	[data alloc_buffer:(int)len];
	[data push:buff Length:(int)len];
	return data;
}
#endif
//-----------------------------------------------------------------------------
//    Z80 CPU Thread
//-----------------------------------------------------------------------------
-(void)z80Thread {
    volatile double clock_sync = system_tick_time() + IVLSEC;
    for (unsigned int tk = 0;; ++tk) {
        int expected = 1;
        if (atomic_compare_exchange_strong(&self->request_cpu_signal, &expected, 9)) {
            NSLog(@"Accept Request CPU HALT");
            self.running = false;
            atomic_store(&self->request_cpu_signal, 0);
            return;
        }
        expected = 2;
        if (atomic_compare_exchange_strong(&self->request_cpu_signal, &expected, 9)) {
            printf("Accept Request CPU RESET\n");
            [self cpu_reset];
            self->state = 0;
            clock_sync = system_tick_time() + IVLSEC;
            atomic_store(&self->request_cpu_signal, 0);
            continue;
        }
        expected = 3;
        if (atomic_compare_exchange_strong(&self->request_cpu_signal, &expected, 9)) {
            NSLog(@"Accept Request CPU PAUSE");
            self.pause = true;
            [pauseZ80sync lock];
            [pauseZ80sync unlock];
            self.pause = false;
            self->state = 0;
            n80device.dmac._penarty = 0;
            clock_sync = system_tick_time() + IVLSEC;
            atomic_store(&self->request_cpu_signal, 0);
            continue;
        }
        // wait for CPU HALT request
        if (self.terminating || n80lx.cpu.r.x.HALT != 0)
            continue;
        
        // z80 step execute
        if (n80device.dmac._penarty > 0 && (tk & 3) != 3) {
            if (n80device.dmac._penarty > 10) {
                n80lx.cpu.r.x.T  = 10;
                n80device.dmac._penarty -= 10;
            } else {
                n80lx.cpu.r.x.T  = n80device.dmac._penarty;
                n80device.dmac._penarty = 0;
            }
            n80lx.cpu.r.x.W = 0;
        } else {
            // Z80 step execute
            if (n80lx.cpu.step(&n80lx.cpu)) {
                // notify if detect illegal instructions
                if (self.haltcallback != nil) {
                    dispatch_async(dispatch_get_main_queue(), ^{
                        self.haltcallback(n80lx.cpu.r.x.HALT);
                    });
                }
            }
        }
        n80lx.cpu.r.x.T += (n80lx.cpu.r.x.TwM1 + n80lx.cpu.r.x.W + n80lx.cpu.r.x.Wh);
        
        // vrtc operation
        int vrtc = n80device.vrtc.blanking(n80lx.cpu.r.x.T);
        if (vrtc == _VRTC_VBBGN) {
            ;
        } else if (vrtc == _VRTC_VBEND) {
            // draw screen
            [self.n80gr draw_offscreen];
        }
        // beep/pcg ticks
        [self->sound append_sound_ticks:n80lx.cpu.r.x.T];
        
        // cpu clock sync
        self->state += n80lx.cpu.r.x.T;
        if (self->state >= n80device.param.sync_ticks) {
            double present = system_tick_time();
            while (present < clock_sync) {
                double diff = clock_sync - present;
                if (diff >= 0.0005) {    // usleep if more than 0.5ms(500ns)
                    struct timespec req;
                    req.tv_sec = 0;
					req.tv_nsec = diff * 1e9 - 100;	//getclock 100ns分差し引く
                    nanosleep(&req, NULL);
                } else if (diff < 0.00005)    // quit less than 50ns
                    break;
                present = system_tick_time();
            }
            self->state -= n80device.param.sync_ticks;
            clock_sync += IVLSEC;    //20msec
            if ((present = system_tick_time()) >= clock_sync)
                clock_sync = present + IVLSEC;
        }
    }
}

-(void)cpu_reset {
    n80lx.cpu.reset(&n80lx.cpu);
    n80device.reset(&n80device);
    [self->sound cpu_reset];
}
//-----------------------------------------------------------------------------
//    Draw off-screen
//-----------------------------------------------------------------------------
-(void)refleshScreen {
    [self.n80gr draw_offscreen];
}
//-----------------------------------------------------------------------------
//    Key Event Handler
//-----------------------------------------------------------------------------
-(void)KeyClear {
    [pauseZ80sync lock];
    int expected = 0;
    while (!atomic_compare_exchange_strong(&self->request_cpu_signal, &expected, 3)) {
        expected = 0;
    }
    //OSAtomicCompareAndSwap32(0, 3, &n80lx.request_disp_signal);
    
    n80device.clear_keys(&n80device);
    
    [pauseZ80sync unlock];
}

-(void)enterPause {
    if (!self.pause) {
        [pauseZ80sync lock];
        int expected = 0;
        while (!atomic_compare_exchange_strong(&self->request_cpu_signal, &expected, 3)) {
            expected = 0;
        }
        //OSAtomicCompareAndSwap32(0, 3, &n80lx.request_disp_signal);
    }
}

-(void)leavePause:(BOOL)terminating {
    if (self.pause) {
        [pauseZ80sync unlock];
    }
    self.terminating = terminating;
}

@end

