//
//  n80em.m
//
//  Copyright (C) 1993-2019 Hiroshi Kanemoto.
//    Released under the MIT license
//    http://opensource.org/licenses/mit-license.php
//
//     Created:    2012/08/22    H.Kanemoto
//    Modified:    2016/04/05    H.Kanemoto    ported from MaxOS X
//    Modified:    2019/05/12    H.Kanemoto    support METAL
//
#import "n80em.h"
#import "n80graph.h"
//-----------------------------------------------------------------------------
//    Emulator C Functions
//-----------------------------------------------------------------------------
extern N80LX        n80lx;
static const DWORD    SPITZ_COLOR[8] = {0xFF000000,0xFFff0000,0xFF00ff00,0xFFffff00,0xFF0000ff,0xFFff00ff,0xFF00ffff,0xFFffffff};
static const DWORD    SPITZ_COLOR2[8] = {0xFF000000,0xFF7f0000,0xFF007f00,0xFF7f7f00,0xFF00007f,0xFF7f007f,0xFF007f7f,0xFF7f7f7f};
static const int    pitch = 640;

static void         _blt4020(LPDWORD, int line);
static void         _blt4025(LPDWORD, int line);
static void         _blt8020(LPDWORD, int line);
static void         _blt8025(LPDWORD, int line);
static void            bltblank(LPDWORD);

typedef void         (*BLTFNC)(LPDWORD, int);
static BLTFNC        bltfncs[] = {_blt4020,_blt4025,_blt8020,_blt8025};
static BLTFNC        fnc = _blt4020;
static DWORD        cur_pix_rgb, cur_pix_rgb2;
static int            vtbl_vram20[20], vtbl_vram25[25];
static LPCDWORD     second_rgb_colors = SPITZ_COLOR2;
//-----------------------------------------------------------------------------
//    Emulator Objective-C Class
//-----------------------------------------------------------------------------
@implementation n80graph

-(id)initWithN80:(n80em*)n80em {
    if (self = [super init]) {
        self->n80em = n80em;
        
        // lock for n80 vram sync
        self->sync = [[NSRecursiveLock alloc] init];
        [self prepareDotPatterns];
        self->draw_queue = dispatch_queue_create("n80_draw_queue", NULL);
    }
    return self;
}

-(void)dealloc {
    NSLog(@"n80graph dealloc");
    [self reqDispose];
}

-(void)lock {
    [self->sync lock];
}

-(void)unlock {
    [self->sync unlock];
}

-(void)suspend {
    atomic_store(&self->drawsync, 0);
}

-(void)resume {
}

- (void)start {
    atomic_store(&self->drawsync, 0);
}
- (void)reqDispose {
    //dispatch_release(self->draw_queue);
    self->sync = nil;
}
//-----------------------------------------------------------------------------
//    Initialize Emulator
//-----------------------------------------------------------------------------
-(void)prepareDotPatterns {
    // fb用ビットパターンデータを作成
    for (int line = 0; line < 20; ++line)
        vtbl_vram20[line] = pitch * (399 - (line * 20));
    for (int line = 0; line < 25; ++line)
        vtbl_vram25[line] = pitch * (399 - (line * 16));
}
-(void)chg_bltfunc:(int)mode {
    // mode: 0 = 40x20, 1 = 40x25, 2 = 80x20, 3 = 80x25
    fnc = bltfncs[mode];
    self->bltmode = mode;
}
-(void)chg_palette:(bool)interrace {
    second_rgb_colors = interrace ? SPITZ_COLOR2 : SPITZ_COLOR;
}
//-----------------------------------------------------------------------------
//    Draw off-screen
//-----------------------------------------------------------------------------
-(void)draw_offscreen {
    if (self->n80em.running) {
        int expected = 0;
        if (atomic_compare_exchange_strong(&self->drawsync, &expected, 1)) {
            /*dispatch_async(self->draw_queue, ^{
                [self draw_offscreen_core];
                atomic_store(&self->drawsync, 0);
            });*/
            [self draw_offscreen_core];
            atomic_store(&self->drawsync, 0);
        }
    }
}
-(void)draw_offscreen_core {
    // purchase bitmap pointer to off-screen
    self->bitmap = (LPDWORD)self->n80em.bmpimg.bitmap_back;
    
    if ((n80device.crtc._text_visible & 0x06) == 0) {
        for (int i = 0; i < 25; ++i) {
            n80device.vram_sync_enter();
            n80device.attr.cache(i);
            n80device.vram_sync_leave();
            n80device.attr.eval(i);
            fnc(self->bitmap, i);
        }
    } else {
        bltblank(self->bitmap);
    }
    //exchange off-screen and primary-screen
    [self->n80em.bmpimg flip];
}
/**--------------------------------------------------------------------------
 **        draw off-screen bitmap
 **------------------------------------------------------------------------*/
static inline void set_pixel_80(LPDWORD vram) {
    *vram           = cur_pix_rgb;
    *(vram-pitch)   = cur_pix_rgb2;
}
static inline void set_pixel_40(LPDWORD vram) {
    *vram           = cur_pix_rgb;
    *(vram-pitch)   = cur_pix_rgb2;
    *(vram+1)       = cur_pix_rgb;
    *(vram-pitch+1) = cur_pix_rgb2;
}
static inline void reset_pixel_80(LPDWORD vram) {
    *vram           = 0;
    *(vram-pitch)   = 0;
}
static inline void reset_pixel_40(LPDWORD vram) {
    *vram           = 0;
    *(vram-pitch)   = 0;
    *(vram+1)       = 0;
    *(vram-pitch+1) = 0;
}
/**--------------------------------------------------------------------------
 **        process font pattern by attribute
 **------------------------------------------------------------------------*/
static inline BYTE proc_video_pixcel25(BYTE attr, BYTE bit, int r) {
    if (attr & 0x80 && r == 7)    // under line
        bit |= 0xff;
    if (attr & 0x08 && r == 0)    // upper line
        bit |= 0xff;
    if (attr & 0x40)            // secret
        bit = 0;
    if (attr & 0x20)            // reverse
        bit = ~bit;
    return bit;
}
static inline BYTE proc_video_pixcel20(BYTE attr, BYTE bit, int r) {
    if (attr & 0x80 && r == 9)    // under line
        bit |= 0xff;
    if (attr & 0x08 && r == 0)    // upper line
        bit |= 0xff;
    if (attr & 0x40)            // secret
        bit = 0;
    if (attr & 0x20)            // reverse
        bit = ~bit;
    return bit;
}
/**--------------------------------------------------------------------------
 **        draw Text(40x20)
 **------------------------------------------------------------------------*/
static inline LPDWORD draw_font_raster4020(LPDWORD vram, int r, BYTE attr, BYTE fnt) {
    register LPDWORD _vram = (vram - (pitch * 2));
    register BYTE pat = proc_video_pixcel20(attr, fnt, r);

    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);    vram += 2;
    pat <<= 1;
    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);    vram += 2;
    pat <<= 1;
    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);    vram += 2;
    pat <<= 1;
    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);    vram += 2;
    pat <<= 1;
    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);    vram += 2;
    pat <<= 1;
    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);    vram += 2;
    pat <<= 1;
    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);    vram += 2;
    pat <<= 1;
    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);

    return _vram;
}
static void _blt4020(LPDWORD bitmap, int line)
{
    if (line < 20) {
        LPBYTE        txts = &n80device.attr._textcach[0];
        LPBYTE        pacs = &n80device.attr._eval_attr[0];
        int            col, offset = vtbl_vram20[line];//pitch * (399 - (line * 20));

        for (col = 0; col < 80; col += 2) {
            BYTE        ascii = *txts;
            BYTE        attr  = *pacs;
            LPBYTE        fnt   = &n80device.font[ascii*8];
            LPDWORD        vram  = &bitmap[offset + (col * 8)];
            int             color = attr & 0x07;

            cur_pix_rgb  = SPITZ_COLOR[color];
            cur_pix_rgb2 = second_rgb_colors[color];
            
            txts += 2;
            pacs += 2;
            
            if (attr & 0x10)                // graph
                fnt += 2048;
            else if (n80device.pcg_mode)    // pcg
                fnt += 4096;

            vram = draw_font_raster4020(vram, 0, attr, 0);
            vram = draw_font_raster4020(vram, 1, attr, *fnt++);
            vram = draw_font_raster4020(vram, 2, attr, *fnt++);
            vram = draw_font_raster4020(vram, 3, attr, *fnt++);
            vram = draw_font_raster4020(vram, 4, attr, *fnt++);
            vram = draw_font_raster4020(vram, 5, attr, *fnt++);
            vram = draw_font_raster4020(vram, 6, attr, *fnt++);
            vram = draw_font_raster4020(vram, 7, attr, *fnt++);
            vram = draw_font_raster4020(vram, 8, attr, *fnt++);
            vram = draw_font_raster4020(vram, 9, attr, 0);
        }
    }
}
/**--------------------------------------------------------------------------
 **        draw Text(40x25)
 **------------------------------------------------------------------------*/
static inline LPDWORD draw_font_raster4025(LPDWORD vram, int r, BYTE attr, BYTE fnt) {
    register LPDWORD _vram = (vram - (pitch * 2));
    register BYTE pat = proc_video_pixcel25(attr, fnt, r);

    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);    vram += 2;
    pat <<= 1;
    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);    vram += 2;
    pat <<= 1;
    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);    vram += 2;
    pat <<= 1;
    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);    vram += 2;
    pat <<= 1;
    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);    vram += 2;
    pat <<= 1;
    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);    vram += 2;
    pat <<= 1;
    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);    vram += 2;
    pat <<= 1;
    pat & 0x80 ? set_pixel_40(vram) : reset_pixel_40(vram);

    return _vram;
}
static void _blt4025(LPDWORD bitmap, int line)
{
    if (line < 25) {
        LPBYTE        txts = &n80device.attr._textcach[0];
        LPBYTE        pacs = &n80device.attr._eval_attr[0];
        int            col, offset = vtbl_vram25[line];//pitch * (399 - (line * 16));
        
        for (col = 0; col < 80; col += 2) {
            BYTE        ascii = *txts;
            BYTE        attr  = *pacs;
            LPBYTE        fnt   = &n80device.font[ascii*8];
            LPDWORD        vram  = &bitmap[offset + (col * 8)];
            int             color = attr & 0x07;

            cur_pix_rgb  = SPITZ_COLOR[color];
            cur_pix_rgb2 = second_rgb_colors[color];
            
            txts += 2;
            pacs += 2;
            
            if (attr & 0x10)                // graph
                fnt += 2048;
            else if (n80device.pcg_mode)    // pcg
                fnt += 4096;

            vram = draw_font_raster4025(vram, 0, attr, *fnt++);
            vram = draw_font_raster4025(vram, 1, attr, *fnt++);
            vram = draw_font_raster4025(vram, 2, attr, *fnt++);
            vram = draw_font_raster4025(vram, 3, attr, *fnt++);
            vram = draw_font_raster4025(vram, 4, attr, *fnt++);
            vram = draw_font_raster4025(vram, 5, attr, *fnt++);
            vram = draw_font_raster4025(vram, 6, attr, *fnt++);
            vram = draw_font_raster4025(vram, 7, attr, *fnt++);
        }
    }
}
/**--------------------------------------------------------------------------
 **        draw Text(80x20)
 **------------------------------------------------------------------------*/
static inline LPDWORD draw_font_raster8020(LPDWORD vram, int r, BYTE attr, BYTE fnt) {
    register LPDWORD _vram = (vram - (pitch * 2));
    register BYTE pat = proc_video_pixcel20(attr, fnt, r);

    pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
    pat <<= 1;
    pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
    pat <<= 1;
    pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
    pat <<= 1;
    pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
    pat <<= 1;
    pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
    pat <<= 1;
    pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
    pat <<= 1;
    pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
    pat <<= 1;
    pat & 0x80 ? set_pixel_80(vram) : reset_pixel_80(vram);

    return _vram;
}
static void _blt8020(LPDWORD bitmap, int line)
{
    if (line < 20) {
        LPBYTE        txts = &n80device.attr._textcach[0];
        LPBYTE        pacs = &n80device.attr._eval_attr[0];
        int            col, offset = vtbl_vram20[line];//pitch * (399 - (line * 20));
        
        for (col = 0; col < 80; ++col) {
            BYTE        ascii = *txts++;
            BYTE        attr  = *pacs++;
            LPBYTE        fnt   = &n80device.font[ascii*8];
            LPDWORD        vram  = &bitmap[offset + (col * 8)];
            int             color = attr & 0x07;

            cur_pix_rgb  = SPITZ_COLOR[color];
            cur_pix_rgb2 = second_rgb_colors[color];
            
            if (attr & 0x10)                // graph
                fnt += 2048;
            else if (n80device.pcg_mode)    // pcg
                fnt += 4096;

            vram = draw_font_raster8020(vram, 0, attr, 0);
            vram = draw_font_raster8020(vram, 1, attr, *fnt++);
            vram = draw_font_raster8020(vram, 2, attr, *fnt++);
            vram = draw_font_raster8020(vram, 3, attr, *fnt++);
            vram = draw_font_raster8020(vram, 4, attr, *fnt++);
            vram = draw_font_raster8020(vram, 5, attr, *fnt++);
            vram = draw_font_raster8020(vram, 6, attr, *fnt++);
            vram = draw_font_raster8020(vram, 7, attr, *fnt++);
            vram = draw_font_raster8020(vram, 8, attr, *fnt++);
            vram = draw_font_raster8020(vram, 9, attr, 0);
        }
    }
}
/**--------------------------------------------------------------------------
 **        draw Text(80x25)
 **------------------------------------------------------------------------*/
static inline LPDWORD draw_font_raster8025(LPDWORD vram, int r, BYTE attr, BYTE fnt) {
    register LPDWORD _vram = (vram - (pitch * 2));
    register BYTE pat = proc_video_pixcel25(attr, fnt, r);

    pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
    pat <<= 1;
    pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
    pat <<= 1;
    pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
    pat <<= 1;
    pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
    pat <<= 1;
    pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
    pat <<= 1;
    pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
    pat <<= 1;
    pat & 0x80 ? set_pixel_80(vram++) : reset_pixel_80(vram++);
    pat <<= 1;
    pat & 0x80 ? set_pixel_80(vram) : reset_pixel_80(vram);

    return _vram;
}
static void _blt8025(LPDWORD bitmap, int line)
{
    if (line < 25) {
        LPBYTE        txts = &n80device.attr._textcach[0];
        LPBYTE        pacs = &n80device.attr._eval_attr[0];
        int            col, offset = vtbl_vram25[line];//pitch * (399 - (line * 16));

        for (col = 0; col < 80; ++col) {
            BYTE        ascii = *txts++;
            BYTE        attr  = *pacs++;
            LPBYTE        fnt   = &n80device.font[ascii*8];
            LPDWORD        vram  = &bitmap[offset + (col * 8)];
            int             color = attr & 0x07;

            cur_pix_rgb  = SPITZ_COLOR[color];
            cur_pix_rgb2 = second_rgb_colors[color];
            
            if (attr & 0x10)                // graph
                fnt += 2048;
            else if (n80device.pcg_mode)    // pcg
                fnt += 4096;

            vram = draw_font_raster8025(vram, 0, attr, *fnt++);
            vram = draw_font_raster8025(vram, 1, attr, *fnt++);
            vram = draw_font_raster8025(vram, 2, attr, *fnt++);
            vram = draw_font_raster8025(vram, 3, attr, *fnt++);
            vram = draw_font_raster8025(vram, 4, attr, *fnt++);
            vram = draw_font_raster8025(vram, 5, attr, *fnt++);
            vram = draw_font_raster8025(vram, 6, attr, *fnt++);
            vram = draw_font_raster8025(vram, 7, attr, *fnt++);
        }
    }
}
/**--------------------------------------------------------------------------
 **        clear off-screen bitmap
 **------------------------------------------------------------------------*/
static void bltblank(LPDWORD bitmap)
{
    int h = 0, i = 0;
    for (h = 0; h < 20; ++h) {
        int offset = vtbl_vram20[h];
        LPDWORD    vram = &bitmap[offset];
        for (i = 0; i < 640; ++i) {
            vram[i]          = 0;
            vram[i-pitch]    = 0;
            vram[i-pitch*2]  = 0;
            vram[i-pitch*3]  = 0;
            vram[i-pitch*4]  = 0;
            vram[i-pitch*5]  = 0;
            vram[i-pitch*6]  = 0;
            vram[i-pitch*7]  = 0;
            vram[i-pitch*8]  = 0;
            vram[i-pitch*9]  = 0;
            vram[i-pitch*10] = 0;
            vram[i-pitch*11] = 0;
            vram[i-pitch*12] = 0;
            vram[i-pitch*13] = 0;
            vram[i-pitch*14] = 0;
            vram[i-pitch*15] = 0;
            vram[i-pitch*16] = 0;
            vram[i-pitch*17] = 0;
            vram[i-pitch*18] = 0;
            vram[i-pitch*19] = 0;
        }
    }
}

@end

